; Some fred commands to aid in preparing the help file.

; 07/09/96 bill set-view-font -> ed-set-view-font as necessary
; 02/06/92 bill set-window-font -> set-view-font
;               fix-spaces-around-italics

(in-package :ccl)

(let ((*warn-if-redefine* nil))
(defmacro def-fred-command (keystroke function &optional doc-string)
  (if (symbolp function) (setq function `',function))
  (if (and (listp function) (eq (car function) 'lambda))
    (setq function `(function ,function)))
  `(comtab-set-key *comtab* ',keystroke ,function ,doc-string))
)

(def-fred-command (:function #\5)
  (lambda (w) (ed-set-view-font w '("New York" 10 :plain))))
(def-fred-command (:function #\6)
  (lambda (w) (ed-set-view-font w '("Monaco" 9 :plain))))
(def-fred-command (:function #\7)
  (lambda (w) (ed-set-view-font w '("New York" 10 :plain :italic))))
(def-fred-command (:function #\8)
  (lambda (w) (ed-set-view-font w '("Monaco" 9 :plain :bold))))	
(def-fred-command (:function #\9)
  (lambda (w)
    (set-view-font w '("New York" 10 :plain))
    (format w "	[")
    (set-view-font w :italic)
    (format w "Function")
    (set-view-font w :plain)
    (format w "]~%")))

(defmethod ed-refresh-screen ((w fred-mixin))
  (let ((frec (frec w)))
    (with-port (wptr w)
      (%update-resized-lines frec))
    (setf (fr.bmod frec) 0
          (fr.zmod frec) 0)
    (frec-draw-contents frec)))(def-fred-command (:control #\l)
  ed-refresh-screen)
(def-fred-command (:control :meta #\w)
  (lambda (w) 
    (setf (fred-wrap-p w) (not (fred-wrap-p w)))
    (ed-refresh-screen w)))


(defvar *italic-ff* (font-codes :italic))

(defun buffer-italic-p (buf &optional pos)
  (not (eql 0 (logand (buffer-char-font-codes buf pos) *italic-ff*))))

(defvar *monaco-ff*)
(defvar *monaco-ff-mask*)
(defvar *9-ms*)
(defvar *9-ms-mask*)
(multiple-value-setq (*monaco-ff* *9-ms* *monaco-ff-mask* *9-ms-mask*)
  (font-codes '("Monaco" 9)))
(setq *monaco-ff* (logand *monaco-ff* *monaco-ff-mask*)
      *9-ms* (logand *9-ms* *9-ms-mask*))
(declaim (fixnum *monaco-ff* *monaco-ff-mask* *9-ms* *9-ms-mask*))

(defun buffer-monaco-9-p (buf &optional pos)
  (multiple-value-bind (ff ms) (buffer-char-font-codes buf pos)
    (and (eql *monaco-ff* (logand ff *monaco-ff-mask*))
         (eql *9-ms* (logand ms *9-ms-mask*)))))

; spaces before or after monaco 9 become monaco 9.
; spaces after italics become monaco 9.
; spaces before italics become italic.
; all double spaces are removed. 
(defun fix-spaces-around-italics (w)
  (let ((m (make-mark (fred-buffer w) 0)))
    ; remove double spaces
    (loop
      (let ((pos (buffer-forward-search m "  ")))
        (unless pos (return))
        (set-mark m (- pos 2))
        (buffer-delete m (1- pos))))

    (set-mark m 0)
    (loop
      (let ((pos (buffer-forward-search m #\space)))
        (unless pos (return))
        (set-mark m pos)
        (let* ((before-something (< pos (buffer-size m)))
               (before-monaco (and before-something (buffer-monaco-9-p m pos)))
               (before-italics (and before-something (buffer-italic-p m pos)))
               (after-something (> (- pos 2) 0))
               (after-monaco (and after-something (buffer-monaco-9-p m (- pos 2))))
               (after-italics (and after-something
                                   (or (buffer-italic-p m (- pos 2))
                                       (and (member (buffer-char m (- pos 2))
                                                    '(#\, #\.))
                                            (> (- pos 3) 0)
                                            (buffer-italic-p m (- pos 3)))))))
          (when (or before-monaco 
                    (and after-monaco (not before-italics))
                    before-italics
                    after-italics)
            (set-selection-range w (1- pos) pos)
            (window-show-selection w)
            (ed-set-view-font w (if (or before-monaco
                                     (and after-monaco (not before-italics))
                                     after-italics)
                               '("Monaco" 9 :plain)
                               '("New York" 10 :plain :italic)))
            (fred-update w)
            ))))))

(def-fred-command (:control :meta :function #\7)
  #'(lambda (w) (eval-enqueue `(fix-spaces-around-italics ,w))))

#| ; This changes (NAME SPECIALIZER) into NAME.
(defun ed-remove-specialization (w)
  (let ((b (fred-buffer w)))
    (if (or (>= (buffer-position b) (buffer-size b))
            (neq #\( (buffer-char b)))
      (ed-beep)
      (progn
        (ed-delete-char w)
        (ed-forward-sexp w)
        (ed-kill-forward-sexp w)
        (when (eq #\) (buffer-char b))
          (ed-delete-char w))))))
(def-fred-command (:control #\z) ed-remove-specialization)

; This looks for italics followed by a single space and allows changing to two spaces
(defparameter *italic-space-comtab*
  (let ((comtab (make-comtab 'qris-next)))
    (comtab-set-key comtab #\space 'qris-replace)
    (comtab-set-key comtab '(:control #\g) 'qris-done)
    comtab))

(defun query-replace-italic-space (w)
  (if (qris-search w)
    (install-shadowing-comtab *italic-space-comtab* nil)
    (ed-beep)))

(def-fred-command (:meta #\s) query-replace-italic-space)

(defun buffer-char-italic-p (buf &optional pos)
  (logbitp 1 (ash (point-h (buffer-char-font-codes buf pos)) -8)))

(defun qris-search (w)
  (let* ((buf (fred-buffer w))
         (size (buffer-size buf))
         (pos (+ (buffer-position buf) 1))
         temp)
    (loop
      (when (or (null pos) (>= pos size)) (return nil))
      (setq pos (buffer-forward-search buf " " pos t))
      (when (and pos
                 (>= (setq temp (- pos 2)) 0)
                 (neq (buffer-char buf temp) #\space)
                 (buffer-char-italic-p buf temp)
                 (not (memq (buffer-char buf pos) '(#\space #\, #\. #\) #\] #\})))
                 (not (buffer-char-italic-p buf pos)))
        (set-selection-range w pos (1- pos))
        (return pos)))))

(defun double-space-search (w)
  (let* ((buf (fred-buffer w))
         (size (buffer-size buf))
         (pos (+ (buffer-position buf) 1))
         temp)
    (loop
      (when (or (null pos) (>= pos size)) (return nil))
      (setq pos (buffer-forward-search buf "  " pos t))
      (when (and pos
                 (>= (setq temp (- pos 3)) 0)
                 (neq (buffer-char buf temp) #\space)
                 (buffer-char-italic-p buf temp)
                 (not (memq (buffer-char buf pos) '(#\space)))
                 (not (buffer-char-italic-p buf pos)))
        (set-selection-range w pos (- pos 2))
        (let* ((buf (fred-buffer w)))
          (multiple-value-bind (s e) (selection-range w)
            (buffer-delete buf s e))
          (multiple-value-bind (ff ms) (buffer-font-codes buf)
            (apply #'buffer-set-font-codes buf *space-font-codes*)
            (buffer-insert buf " ")
            (buffer-set-font-codes buf ff ms)))
        (return pos)))))

(defun qris-next (w)
  (unless (qris-search w)
    (qris-done w)))

(defun qris-done (w)
  (declare (ignore w))
  (remove-shadowing-comtab "Done."))

(defvar *space-font-codes*
  (multiple-value-bind (ff ms)
                       (font-codes '("Monaco" 9 :plain))
    (list ff ms)))

(defun qris-replace (w)
  (qris-replace-no-search w)
  (qris-next  w))

(defun qris-replace-no-search (w)
  (let* ((buf (fred-buffer w)))
    (multiple-value-bind (s e) (selection-range w)
      (buffer-delete buf s e))
    (multiple-value-bind (ff ms) (buffer-font-codes buf)
      (apply #'buffer-set-font-codes buf *space-font-codes*)
      (buffer-insert buf " ")
      (buffer-set-font-codes buf ff ms))))

(defun qris-replace-all (w)
  (while (qris-search w)
    (qris-replace-no-search w)))

|#

#|
(setq *old-tab* (comtab-get-key *comtab* #\tab))
(def-fred-command #\tab
  ed-self-insert)
|#
;  #.*old-tab*)
