;;; Traps, foreign-function calls, callbacks.

	include 'lisp.s'

	import jmpsym
;;; 05/18/02 akh 16 byte align stack in ffcallslep
;;; 05/11/01 akh see comment re crock in ffcalladdress_proper or some such
;;; 03/26/01 akh use r12 vice r11 they tell me for OSX  - in ffcalladdress_proper
;;; 02/17/01 akh see bypass in ffcalladdress_proper - OSX stuff
;;; 02/11/97 gb    maintain fp_s32conv.
;;; 10/11/96 bill  callback saves the C preserved registers on the SP stack instead of the TSP stack.
;;;                This prevents TSP from overflowing while a VSP overflow is calling back.
;;; -------------  4.0f1
;;; 10/05/96 bill  ffcallslep saves floating point registers around its call to resolve-slep-address
;;; -------------  4.0b2
;;; 07/16/96 bill  Back out of storing VSP at entry and after pushing saved regs.
;;;                Since its savelr is 0, the stack overflow code will skip
;;;                the ffcall frame in determining how much stack to copy.
;;;                This also means we don't need to save the VSP in the ffcall stack frame.
;;;                Store zero at c_frame.savelr(sp) before vpush'ing the saved registers.
;;;                This ensures that the VSP stack overflow code will skip this frame.
;;; 07/08/96 bill  callback stores VSP in dummy stack frame for stack overflow code.
;;; 07/06/96 bill  callback stores nil-1 in rnil before pushing a TSP frame,
;;;                so that the TSP overflow code can do a consistency check.
;;;                ffcall stores vsp at lisp_frame.savevsp(sp) before pushing
;;;                to make the stack overflow code happy.
;;; 07/04/96 bill  breakpoint subprim
;;; 06/18/96 bill  ffcall remembers both VSP at entry and VSP after pushing saved regs
;;;                (may overflow the current VSP segment while pushing the saved regs).
;;;                callback/@no_exception restores from 4(vsp), not (vsp)
;;; 06/13/96 gb    maintain fp_zero reg, clear more fpscr status bits.
;;; --- 3.9 release ---
;;; 04/08/96 bill  callback initializes saved registers from either the
;;;                exception_saved_registers where callback_to_lisp saved them
;;;                or the VSP where ffcall saved them.
;;;                callback uses imm3=r6 for the savelr instead of r5.
;;;                The latter got overwritten before its second use.
;;; 03/28/96 gb    ffcalladdress: set ffi_exception register botch
;;; 03/18/96 gb    hidden foreign areas; ffcalladdress_proper
;;;                maintains current_cs.active.
;;; 03/14/96 bill  ffcallslep
;;; 03/07/96 gb    muck with fpscr
;;; 11/27/95 bill  callback sets save_vsp & save_tsp globals on exit.
;;; 11/11/95 bill  callback
;;;                ff_calladdr saves vsp in the save_vsp global
;;; 11/07/95  gb  regtrap uses only low 8 bits to determine trap address.


; Enter the debugger
breakpoint: spentry breakpoint
	li r3,0
	tw 28,sp,sp		; 28 = lt|gt|eq (assembler bug for the latter)
	blr			; if handler didn't


;;; Call the UPP corresponding to the (68k-register-based) trap number in
;;; arg_z.
	
regtrap: spentry regtrap
	unbox_fixnum imm1,arg_z
	andi. arg_z,arg_z,0xff<<fixnumshift
	ref_global imm0,os_trap_base
	lwzx imm0,imm0,arg_z
	ref_global arg_z,os_trap_call
	stw imm0,c_frame.params+(0<<2)(sp)
	stw imm1,c_frame.params+(2<<2)(sp)
	b ffcalladdress
	

;;; Call the UPP corresponding to the (68k-stack-based) trap number in
;;; arg_z.
strap:	spentry strap
	andi. arg_z,arg_z,0x3ff<<fixnumshift
	ref_global imm0,tb_trap_base
	lwzx imm0,imm0,arg_z
	ref_global arg_z,tb_trap_call
	stw imm0,c_frame.params+(0<<2)(sp)
	b ffcalladdress
	
;;; This subprim is used by the trap expansion code.
;;; It takes a shared-library-entry-point (slep) instance in arg_z and
;;; the other arguments as for ffcall.
;;; If (slep.address slep) is non-nil, it is the (fixnum) address of the function to call.
;;; Otherwise, call back to resolve-slep-address to look up the address.
ffcallslep: spentry ffcallslep
	mr arg_y,arg_z
	vref32 arg_z,arg_z,1                   ; (slep.address slep)
	cmpw cr0,arg_z,rnil
	bne cr0,ffcalladdress

	;; this matches the code between ffcalladdress & ffcalladdress_proper
	mflr loc_pc
	stw rzero,c_frame.savelr(sp)           ; make sure stack overflow code skips this frame
	vpush_saveregs
	lwz imm1,0(sp)
	la imm2,-lisp_frame.size(imm1)
	stw imm1,0(imm2)
	stw imm2,0(sp)
	stw loc_pc,lisp_frame.savelr(imm2)
	stw vsp,lisp_frame.savevsp(imm2)
	stw fn,lisp_frame.savefn(imm2)

	;; Save the floating point registers in a new stack frame
	;; Can't use temp stack here (unless we pushed a catch frame),
	;; since apply-in-frame wouldn't know about the extra temp stack frame.
	stwu imm2,-(13*8 + 8 + lisp_frame.size)(sp)  ; round up to 16 byte align
	stw rzero,lisp_frame.savelr(sp)
	la imm1,lisp_frame.size-8(sp)
	stfdu fp1,8(imm1)
	stfdu fp2,8(imm1)
	stfdu fp3,8(imm1)
	stfdu fp4,8(imm1)
	stfdu fp5,8(imm1)
	stfdu fp6,8(imm1)
	stfdu fp7,8(imm1)
	stfdu fp8,8(imm1)
	stfdu fp9,8(imm1)
	stfdu fp10,8(imm1)
	stfdu fp11,8(imm1)
	stfdu fp12,8(imm1)
	stfdu fp13,8(imm1)

	;; Call resolve-slep-address to get the address (or error)
	li fn,0
	mr arg_z,arg_y
	la fname,nrs.resolve_slep_address(rnil)
	set_nargs 1
	bl jmpsym

	;; Restore the floating point registers and pop off the new stack frame
	la imm1,lisp_frame.size-8(sp)
	lfdu fp1,8(imm1)
	lfdu fp2,8(imm1)
	lfdu fp3,8(imm1)
	lfdu fp4,8(imm1)
	lfdu fp5,8(imm1)
	lfdu fp6,8(imm1)
	lfdu fp7,8(imm1)
	lfdu fp8,8(imm1)
	lfdu fp9,8(imm1)
	lfdu fp10,8(imm1)
	lfdu fp11,8(imm1)
	lfdu fp12,8(imm1)
	lfdu fp13,8(imm1)
	la sp,(13*8 + 8 + lisp_frame.size)(sp)

	;; Finally, continue with the ff-call
	lwz imm2,c_frame.backlink(sp)
	b ffcalladdress_proper

;;; Caller has created a C frame on SP, with 4 "extra" fullwords at the bottom
;;; of the frame.  C-style arguments are sitting in that frame (we load all
;;; of the C argument registers); a macptr to the function descriptor is in
;;; arg_z.
;;; We save all of the lisp NVRs and other state, use the 4 extra words to
;;; make a lisp frame, do the ff-call, discard the C frame, recover the
;;; saved registers and other state, and return to the lisp frame (with the
;;; result in imm0 (R3)).
ffcall:	spentry ffcall
	lwz arg_z,macptr.address(arg_z); a fixnum
	endf
ffcalladdress:	spentry ffcalladdress
	mflr loc_pc
	stw rzero,c_frame.savelr(sp)           ; make sure stack overflow code skips this frame
	vpush_saveregs
	lwz imm1,0(sp)
	la imm2,-lisp_frame.size(imm1)
	stw imm1,0(imm2)
	stw imm2,0(sp)
	stw loc_pc,lisp_frame.savelr(imm2)
	stw vsp,lisp_frame.savevsp(imm2)
	stw fn,lisp_frame.savefn(imm2)
ffcalladdress_proper:
	;; imm2 = lisp_frame (under c_frame on tos.)
	ref_global imm3,current_cs
	stw imm2,area.active(imm3)
	set_global freeptr,save_freeptr
	set_global tsp,save_tsp
	set_global vsp,save_vsp                ; where callback can find it
	mr r12,arg_z                           ; use r12 vice r11 they tell me for OSX
	set_global rzero,ffi_exception
        lfs fp_zero,nrs.globals.short_float_zero(rnil)		; ensure that fp_zero contains 0.0 should not be needed here
	mtfsf 0xfc,fp_zero
	stw rnil,c_frame.savetoc(sp)
	;; Could use lswi, but that's frowned on.
	lwz r3,c_frame.params+(0*4)(sp)
	lwz r4,c_frame.params+(1*4)(sp)
	lwz r5,c_frame.params+(2*4)(sp)
	lwz r6,c_frame.params+(3*4)(sp)
	lwz r7,c_frame.params+(4*4)(sp)
	lwz r8,c_frame.params+(5*4)(sp)
	lwz r9,c_frame.params+(6*4)(sp)
	lwz r10,c_frame.params+(7*4)(sp)
	lwz r0,0(r12)
	mtctr r0
	lwz rtoc,4(r12)
	;; If anybody cares what the emulator returns in d0, don't
	;; step through any of this code (until we've grabbed the result),
	;; since the debugger may run emulated and clobber it.
        ;; this is at 0x0008c918 - 4 under OSX FC
	bctrl 
	lwz imm2,c_frame.savetoc(sp); recover rnil
	lwz imm1,nrs.globals.emulator_registers(imm2)        
        cmpwi imm1,0
        mr arg_z,imm2
       ;; if osx we dont want to reference location  0
        beq bypass
	lwz imm1,0(imm1)
	;; Ok to step now
	extsh imm1,imm1
	box_fixnum arg_z,imm1
bypass:	lwz sp,0(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	vpop_saveregs
	li rzero,0
	mr loc_g,rzero
	mr arg_x,imm2
	mr arg_y,imm2
#	mr arg_z,imm2
	mr temp0,imm2
	mr temp1,imm2
	mr temp2,imm2
	mr temp3,imm2
	mr fn,imm2
	mr rnil,imm2
       ;; should not be needed but may fix a crock in Classic and OSX??
        lfs fp_zero,nrs.globals.short_float_zero(rnil)		; ensure that fp_zero contains 0.0
	ref_global freeptr,save_freeptr
	mr initptr,freeptr
	ref_global tsp,save_tsp
	lwz loc_pc,lisp_frame.savelr(sp)
	mtlr loc_pc
	lwz fn,lisp_frame.savefn(sp)
	la sp,lisp_frame.size(sp)
	cmpwi cr0,arg_z,0
	mcrfs cr1,0
	li imm2,0		; assume no FP exception
	set_global imm2,ffi_exception
	blelr cr1
	lwi imm2,0x40000000
	set_global imm2,ffi_exception
	bns cr1,@no_ox
; The mcrfs cleared OX.  Set it again, then clear FEX.
	mtfsb1 3
@no_ox:	mtfsb0 1
	blr

;;; On entry, TOC -> nilreg, callback-index
;;; Restore lisp context, then funcall #'%pascal-functions% with
;;; two args: callback-index, args-ptr (a macptr pointing to the args on the stack)
;;; When it returns, coerce the result, either a macptr or an integer, to 32 bits
;;; and return as a C function would.
callback:       spentry callback
	;; Save arg registers (r3-r10)
	stw r3,c_frame.params+(0*4)(sp)
	stw r4,c_frame.params+(1*4)(sp)
	stw r5,c_frame.params+(2*4)(sp)
	stw r6,c_frame.params+(3*4)(sp)
	stw r7,c_frame.params+(4*4)(sp)
	stw r8,c_frame.params+(5*4)(sp)
	stw r9,c_frame.params+(6*4)(sp)
	stw r10,c_frame.params+(7*4)(sp)
	mflr imm3
	stw imm3,c_frame.savelr(sp)
	mfcr imm0
	stw imm0,c_frame.crsave(sp)

	;; This code depends on the fact that imm0, imm1, imm3 & tsp are
	;; not r13-r31 (the non-volatile C registers).
	lwz imm0,4(rtoc)                ; callback-index
	lwz imm1,0(rtoc)                ; nil
	la rnil,-1(imm1)                ; Marker for TSP overflow code
	lwz tsp,nrs.globals.save_tsp(imm1)

	;; Save the non-volatile registers on the sp stack
	;; This is a non-standard stack frame, but noone will ever see it,
        ;; so it doesn't matter. It will look like more of the stack frame pushed below.
	stwu sp,-(c_reg_save.size)(sp)
	;stw sp,c_reg_save.tsp_mark(sp)                  ; "raw" tsp frame.
	stmw r13,c_reg_save.save_gprs(sp)
	stfd fp_zero,c_reg_save.save_fp_zero(sp)
	stfd fp_s32conv,c_reg_save.save_fp_s32conv(sp)
	mffs fp0
	stfd fp0,-8(sp)
	lwz r31,-4(sp)
	lwz r30,nrs.globals.ffi_exception(imm1)
	or r31,r30,r31
	stw r31,c_reg_save.save_fpscr(sp)
	lwi r30,0x43300000
	lwi r31,0x80000000
	stw r30,-8(sp)
	stw r31,-4(sp)
	lfd fp_s32conv,-8(sp)
	lfs fp_zero,nrs.globals.short_float_zero(imm1)		; ensure that fp_zero contains 0.0
	mtfsf 0xfc,fp_zero		; clear "status", preserve "control"

	;; Restore rest of Lisp context.
	;; Could spread out the memory references here to gain a little speed
	lwz freeptr,nrs.globals.save_freeptr(imm1)
	mr initptr,freeptr
	lwz vsp,nrs.globals.save_vsp(imm1)
	li rzero,0
	mtxer rzero                     ; eventually, lisp may depend on the overflow bit being clear
	mr loc_g,rzero
	mr loc_pc,rzero
	mr fn,rzero                     ; subprim, not a lisp function
	mr temp3,imm1
	mr temp2,imm1
	mr temp1,imm1
	mr temp0,imm1
	mr arg_x,imm1

	; restore saved registers.
	; They are in exception_saved_registers if it is non-zero.
	; Otherwise, they are on top of the VSP stack (but 1)
	lwz imm2,nrs.globals.exception_saved_registers(imm1)
	cmpwi cr0,imm2,0
	beq cr0,@no_exception
	stw rzero,nrs.globals.exception_saved_registers(imm1)
	b @restore_regs
@no_exception:
	la imm2,4(vsp)
@restore_regs:
	restore_saveregs imm2

	; load the arg registers & nargs and callback to the lisp
	mr arg_y,imm0                   ; callback-index
	la arg_z,c_reg_save.size+c_frame.params(sp)     ; parameters (tagged as a fixnum)
	set_nargs 2
	mr rnil,imm1                    ; lisp context restored
	ref_global imm2,current_cs
	lwz imm4,area.active(imm2)
	stwu imm4,-lisp_frame.size(sp)
	stw imm3,lisp_frame.savelr(sp)
	stw vsp,lisp_frame.savevsp(sp)	; for stack overflow code
	la fname,nrs.callbacks(rnil)    ; %pascal-functions%
	bl jmpsym
	lwz imm2,lisp_frame.backlink(sp)
	ref_global imm3,current_cs
	stw imm2,area.active(imm3)
	la sp,lisp_frame.size(sp)
	;; save_freeptr & save_memo may be different due to consing
	;; save_vsp will be restored from ff_call's stack frame, but
	;; I included it here for consistency.
	;; save_tsp is set below after we exit Lisp context.
	set_global freeptr,save_freeptr
	set_global vsp,save_vsp

	;; Exit lisp context
	;; This is not necessary yet, but will be once we can be interrupted
	mr imm1,rnil
	li rnil,0                       ; could lwz rnil,c_reg_save.size+c_frame.savetoc(sp), but that costs time

	;; Restore the non-volatile registers & fpscr
	lwz r31,c_reg_save.save_fpscr(sp)
	rlwinm r30,r31,0,1,1	; extract FEX bit
	rlwinm r31,r31,0,2,0	; clear it
	stw r31,-4(sp)
	lfd fp0,-8(sp)
	mtfsf 0xff,fp0
	stw r30,nrs.globals.ffi_exception(imm1)
	lmw r13,c_reg_save.save_gprs(sp)
	lfd fp_zero,c_reg_save.save_fp_zero(sp)
	lfd fp_s32conv,c_reg_save.save_fp_s32conv(sp)
	lwz sp,0(sp)
	stw tsp,nrs.globals.save_tsp(imm1)

	;; The lisp code has left the return value in c_frame.params(sp)
	;; We just need to load it into R3 to return to the caller
	lwz r3,c_frame.params(sp)

	lwz r4,c_frame.crsave(sp)
	mtcr r4
	lwz r4,c_frame.savelr(sp)
	mtlr r4
	blr

	end